﻿#include <iostream>
#include <list>
#include <string>
#include <map>

using namespace std;

/**
 * \brief invoice, which contains number, amount, month of creation and category.
 */
class invoice
{
private:
    int number_ = 0;
    double amount_;
    unsigned short month_ = 0;
    int category_ = 0;
public:
    /**
     * \brief collection of categories
     */
    static const char* categories[];
    /**
     * \brief collection of months
     */
    static const char* months[];

    /**
     * \brief Creates a new instance of invoice
     */
    invoice(const int number, const double amount, const unsigned short month, const int category) : number_(number),
        amount_(amount),
        month_(month),
        category_(category)
    {
    }

    /**
     * \brief Get number of invoice.
     * \return current number
     */
    int get_number() const
    {
        return number_;
    }

    /**
     * \brief get month of invoice
     * \return month as index
     */
    unsigned short get_month() const
    {
        return month_;
    }

    /**
     * \brief get category of invoice
     * \return category as index
     */
    int get_category() const
    {
        return category_;
    }

    /**
     * \brief get amount of invoice
     * \return amount in euro
     */
    double get_amount() const
    {
        return amount_;
    }

    /**
     * \brief Implements the equality operator.
    */
    bool operator ==(const invoice& e) const
    {
        return number_ == e.number_;
    }

    /**
     * \brief Implements the inequality operator.
     */
    bool operator !=(const invoice& e) const { return !operator==(e); }
};

const char* invoice::categories[] = {
    "Tanken",
    "Nahrungsmittel",
    "Reinigungsmittel",
    "Urlaub",
    "Haustiere",
    "Bekleidung",
    "Wohnen",
    "Versicherung"
};

const char* invoice::months[] = {
    "Januar",
    "Februar",
    "M\x84rz",
    "April",
    "Mai",
    "Juni",
    "Juli",
    "August",
    "September",
    "Oktober",
    "November",
    "Dezember"
};

/**
 * \brief ask a question and get a answer.
 * \tparam T return parameter
 * \param question the question.
 * \return the value.
 */
template <typename T>
T ask_question(const string& question)
{
    cout << question << endl;
    cout << "> ";
    T value;
    cin >> value;
    return value;
}

void create_invoice();
int get_month();
int get_category();
void invoices_ordered_by_month();
void delete_invoice();
void sum_invoices_per_category();

/**
 * \brief compares months ascending.
 * \param first the first invoice
 * \param second the second invoice
 * \return true, if first is smaller than second. False otherwise.
 */
bool compare_month_ascending(const invoice& first, const invoice& second)
{
    return first.get_month() < second.get_month();
}

/**
 * \brief Double linked list of all invoices.
 */
list<invoice> invoices;
/**
 * \brief Consecutive invoice number
 */
int invoice_serial_number;

/**
 * \brief Start point of application
 */
int main()
{
    while (true)
    {
        system("cls");
        cout << "Ausgabenverwaltung (" << invoices.size() << " Rechnungen)" << endl << endl;
        cout << "Bitte w\x84hlen Sie eine Option:" << endl;
        cout << "1: Erfassung einer Rechnung (inkl. Abfrage der Rechnungsdaten)" << endl;
        cout << "2: Alle Ausgaben, sortiert nach Monat, ausgeben" << endl;
        cout << "3: Summe der Ausgaben pro Kategorie ausgeben" << endl;
        cout << "4: Rechnung l\x94schen" << endl;
        cout << endl;
        cout << "0: Programm beenden" << endl;
        const auto input = ask_question<int>("Ihre Eingabe:");

        switch (input)
        {
        case 1:
            create_invoice();
            break;
        case 2:
            invoices_ordered_by_month();
            break;
        case 3:
            sum_invoices_per_category();
            break;
        case 4:
            delete_invoice();
            break;
        case 0:
            return 0;
        default:
            cout << "Unerlaubte Eingabe" << endl;
            system("pause");
        }
    }
}

void create_invoice()
{
    system("cls");
    cout << "Eine neue Rechnung erstellen" << endl;
    cout << "------------------------------------" << endl;
    const auto invoice_no = ++invoice_serial_number;
    const auto amount = ask_question<double>("Bitte geben Sie den Betrag in Euro ein.");
    const auto month = get_month();
    const auto category = get_category();
    const auto i = invoice(invoice_no, amount, month, category);
    invoices.push_back(i);
}

void invoices_ordered_by_month()
{
    system("cls");
    cout << "Alle Rechnungen nach Monat sortiert" << endl;
    cout << "------------------------------------" << endl;
    //<month, <category, amount>>
    map<int, map<int, double>> month_category_amount;
    for (auto invoice : invoices)
    {
        month_category_amount[invoice.get_month()][invoice.get_category()] += invoice.get_amount();
    }

    for (auto month_pair : month_category_amount)
    {
        auto sum = 0.0;
        cout << invoice::months[month_pair.first] << endl;
        cout << "------------------------------------" << endl;
        for (const auto category_amount : month_pair.second)
        {
            cout << invoice::categories[category_amount.first] << ": " << category_amount.second << " Euro" << endl;
            sum += category_amount.second;
        }
        cout << endl;
        cout << "Gesamt: " << sum << " Euro";
        if (sum > 500) cout << "  Mehr ausgegeben als geplant!!!";
        cout << endl;
        cout << "------------------------------------" << endl;
    }
    system("pause");
}

void delete_invoice()
{
    system("cls");
    cout << "Rechnung l\x94schen" << endl;
    cout << "------------------------------------" << endl;
    const auto invoice_number = ask_question<int>("Bitte geben Sie eine Rechnungsnummer ein");
    for (auto invoice : invoices)
    {
        if (invoice.get_number() == invoice_number)
        {
            invoices.remove(invoice);
            return;
        }
    }
    cerr << "Fehler: Rechnung nicht gefunden." << endl;
    system("pause");
}

void sum_invoices_per_category()
{
    system("cls");
    cout << "Summe der Ausgaben pro Kategorie ausgeben" << endl;
    cout << "------------------------------------" << endl;
    map<int, double> sums;
    for (auto invoice : invoices)
    {
        sums[invoice.get_category()] = sums[invoice.get_category()] += invoice.get_amount();
    }
    for (const auto sum : sums)
    {
        cout << "  Kategorie: " << invoice::categories[sum.first] << endl;
        cout << "  Summe: " << sum.second << " Euro" << endl;
        cout << "------------------------------------" << endl;
    }
    system("pause");
}

int get_month()
{
    while (true)
    {
        cout << "Bitte w\x84hlen sie einen Monat aus:" << endl;
        auto index = 0;
        for (auto month : invoice::months)
        {
            cout << ++index << ": " << month << endl;
        }
        const auto answer = ask_question<unsigned int>("");
        const auto size = static_cast<int>(sizeof(invoice::months) / sizeof(*invoice::months));
        if (answer > 0 && answer <= size)
        {
            return answer - 1;
        }
        cout << "Ung\x81ltige Eingabe" << endl << endl;
    }
}

int get_category()
{
    while (true)
    {
        cout << "Bitte w\x84hlen sie eine Kategorie aus:" << endl;
        auto index = 0;
        for (auto category : invoice::categories)
        {
            cout << ++index << ": " << category << endl;
        }
        const auto answer = ask_question<int>("");
        const auto size = static_cast<int>(sizeof(invoice::categories) / sizeof(*invoice::categories));
        if (answer > 0 && answer <= size)
        {
            return answer - 1;
        }
        cout << "Ung\x81ltige Eingabe" << endl << endl;
    }
}